﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace cavfill
{
    class Program
    {

        public static string usage = @"Usage: cavfill [-i bytes] [-s size] [-b buffersize] target_file

Options:
    -s size         Ignore any cavities smaller than _size_ bytes
                    (default size: 50)
    -b buffersize   Leave _buffersize_ bytes alone at edges of cavity
                    (default buffersize: 8)
    -i bytes        Ignore any cavities in the first _bytes_ bytes
                    (default ignore: 0)
    -v              Print version information and exit
    -?              Get help

cavfill processes a target .exe, filling 'cavities' (long sequences of the 
zero byte) with random bytes.  The output is placed in 'cavfill.out'.

The purpose is to serve as a post-compilation step, to prevent malware from
being embedded in these empty (and normally useless) spaces.

The assumption is that if the zero bytes were unnecessary to the execution
of the program it should still work correctly.  (NO GUARANTEES!)

This program is experimental.  I make no guarantees as to the usefulness
of the output.  It may very well make a version of your .exe that appears
to work but has some random bug.  It might even (the fill is random after
all) result in a working output sometimes and a broken output other times.

Seriously.  If you are screwing around with an executable post compliation
and break it somehow, don't come whining to me.

On the other hand I would imagine that this is *relatively* safe, since
Malware hijacks these cavities all the time and usually gets away with it.

";

        public static string version = @"cavfill.exe 0.1 Mar 3 2016
Please see http://www.elifulkerson.com for updates.

";

        static void Main(string[] args)
        {
            if (args.Count() == 0)
            {
                Console.Write(cavfill.Program.usage);
                Environment.Exit(0);
            }

            string path = args[0];
            int cavesize = 50;   // this is the size of the cavity we are looking for
            int zerobuffer = 8;  // When we find a void, we leave 'zerobuffer' 0's alone on each side of it instead of going edge-to-edge in case those 0's are important.
            int ignorebytes = 0;

            for (int i = 0; i < args.Count(); i++)
            {
                path = args[i];

                if (args[i] == "-s" || args[i] == "/s")
                {
                    try
                    {
                        cavesize = Convert.ToInt32(args[i + 1]);
                    }
                    catch { }
                    i++;
                }

                if (args[i] == "-b" || args[i] == "/b")
                {
                    try
                    {
                        zerobuffer = Convert.ToInt32(args[i + 1]);
                    }
                    catch { }
                    i++;
                }

                if (args[i] == "-i" || args[i] == "/i")
                {
                    try
                    {
                        ignorebytes = Convert.ToInt32(args[i + 1]);
                    }
                    catch { }
                    i++;
                }

                if (args[i] == "-v" || args[i] == "/v")
                {
                    Console.Write(cavfill.Program.version);
                    Environment.Exit(1);
                }

                if (args[i] == "-?" || args[i] == "/?")
                {
                    Console.Write(cavfill.Program.usage);
                    Environment.Exit(1);
                }
            }


            
            byte[] file = null;
            try
            {
                file = File.ReadAllBytes(path);
            }
            catch (Exception e)
            {
                Console.WriteLine("Input Error: {0}", e.Message);
                Environment.Exit(1);
            }

            FileStream outfile = null;
            try
            {
                outfile = new FileStream("cavfill.out", FileMode.CreateNew, FileAccess.Write);
            }
            catch (Exception e)
            {
                Console.WriteLine("Output Error: {0}",e.Message);
                Environment.Exit(1);
            }
           
            
            using (MemoryStream memory = new MemoryStream(file))
            {
                
                using (BinaryReader reader = new BinaryReader(memory))
                {
                

                    int foundstart = -1;
                    int foundend = -1;

                    Random random = new Random();

                    byte rbyte;

                    for (int i = 0; i < file.Length; i++)
                    {

                        byte result = reader.ReadByte();

                        if (result == 0 && ignorebytes < i)
                        {
                            if (foundstart == -1) { foundstart = i; }
                        }
                        else
                        {
                            if (foundstart != -1)
                            {
                                foundend = i - 1;

                                if (foundend - foundstart > cavesize)
                                {
                                    Console.WriteLine("Found Cavity: {0} to {1}, size: {2}", foundstart.ToString("X").PadLeft(8, '0'), foundend.ToString("X").PadLeft(8, '0'), foundend - foundstart);

                                    // fill it with rubble

                                    for (int j = foundstart; j <= foundend; j++)
                                    {
                                        // make this random later, but for now FF

                                        if (j - foundstart < zerobuffer || j > foundend - zerobuffer)
                                        {
                                            outfile.WriteByte(0);
                                        }
                                        else
                                        {
                                            //outfile.WriteByte(255);
                                            rbyte = Convert.ToByte(random.Next(256));
                                            outfile.WriteByte(rbyte);

                                        }
                                    }

                                } else {

                                    // not big enough to fill
                                    //Console.WriteLine("Found Small Void: {0} to {1}, size: {2}", foundstart.ToString("X").PadLeft(8, '0'), foundend.ToString("X").PadLeft(8, '0'), foundend - foundstart);
                                    for (int j = foundstart; j <= foundend; j++) {
                                        //Console.WriteLine("filling in");
                                        outfile.WriteByte(0);
                                    }

                                }             

                                foundstart = -1;
                            }
                            // finally, write our non-zero byte always
                            outfile.WriteByte(result);
                        }
                    }

                    // in case we had trailing zeros
                    if (foundstart != -1)
                    {
                        foundend = file.Length -1;

                        if (foundend - foundstart > cavesize)
                        {
                            Console.WriteLine("Found Cavity: {0} to {1} (END), size: {2}", foundstart.ToString("X").PadLeft(8, '0'), foundend.ToString("X").PadLeft(8, '0'), foundend - foundstart);

                            for (int j = foundstart; j <= foundend; j++)
                            {
                                if (j - foundstart < zerobuffer || j > foundend - zerobuffer)
                                {
                                    outfile.WriteByte(0);
                                }
                                else
                                {
                                    //outfile.WriteByte(255);
                                    rbyte = Convert.ToByte(random.Next(256));
                                    outfile.WriteByte(rbyte);
                                }
                            }
                        }
                        else
                        {
                            // not big enough to fill
                            for (int j = foundstart; j <= foundend; j++)
                            {
                                outfile.WriteByte(0);
                            }
                        }
                    }
                }
            }
        }
    }
}
